local InFlight = InFlight
local self = InFlight

local smed = AceLibrary("SharedMedia-1.0")
local dew = AceLibrary("Dewdrop-2.0")

local vars, db, opts				-- addon
local source, destination			-- locations
local endTime, totalTime				-- time
local porttaken, initbulk			-- flags
local rat, iloffset					-- cache variables
local f, spark, timeText, locText, bord	-- frame variables
local gtt = GameTooltip

local locformat = "%s --> %s"

-- called from InFlight_Load
----------------------------
function InFlight:LoadBulk()
----------------------------
	self:RegisterDB("InFlightDB")
	self:RegisterDefaults("profile", {
		confirmflight = false,
		dismount = true, 
		fill = "Fill up",
		inline = false,
		border = "Default",
		height = 14,
		width = 230,
		fontsize = 12,
		texture = "smooth",
		barcolor = { r = 0.5, g = 0.5, b = 0.8, },
		unknowncolor = { r = 0.2, g = 0.2, b = 0.4, },
		bordercolor = { r = 0.6, g = 0.6, b = 0.6, },
		backcolor = { r = 0.1, g = 0.1, b = 0.1, },
		fontcolor = { r = 1.0, g = 1.0, b = 1.0, },
	} )
	db = self.db.profile
	db.poidata = nil
	self:LoadOptions()
	smed:Register("statusbar", "Smooth", "Interface\\AddOns\\InFlight\\smooth")
	
	self:RegisterEvent("TAXIMAP_OPENED")
	self:RegisterEvent("SharedMedia_SetGlobal", "UpdateLook")
	
	if TaxiFrame:IsShown() then
		self:TAXIMAP_OPENED()
	end
	
	self.loaded = true
end

-----------------------------------
function InFlight:OnProfileEnable()
-----------------------------------
	db = self.db.profile
end

-- simple time format
local mod = math.fmod
local function FormatTime(secs)
	return format("%d:%02d", secs / 60, mod(secs, 60))
end

-- shorten name to lighten saved vars
local function ShortenName(name)
	local found = strfind(name, ", ")
	if found then 
		name = strsub(name, 1, found - 1) 
	end
	return name
end

-- setup only when the taxi map is opened the first time
--------------------------------------
function InFlight:TAXIMAP_OPENED(misc)
--------------------------------------
	if not initbulk then
		if not InFlightVars then 
			InFlightVars = (self.LoadDefaults and self:LoadDefaults()) or { Alliance = {}, Horde = {}, }
		end
		if not db.TBCLoaded and self.LoadTBCData then
			self:LoadTBCData()
			db.TBCLoaded = 1
		end
		vars = InFlightVars[UnitFactionGroup("player")]
	
		self:Hook("TakeTaxiNode", true)
		self:SecureHook("TaxiNodeOnButtonEnter")
		
		-- function hooks to detect if a user took a summon
		self:SecureHook("AcceptBattlefieldPort", function(index,accept) porttaken = accept and true end)
		self:SecureHook("ConfirmSummon", function() porttaken = true end)
		
		initbulk = true
	end
	if not misc then
		for i = 1, NumTaxiNodes(), 1 do
			if TaxiNodeGetType(i) == "CURRENT" then
				source = ShortenName(TaxiNodeName(i))
				break
			end
		end
	end
	if db.dismount then 
		Dismount() 
	end
	if Cartographer_Notes then 
		self:NoteIt(misc) 
	end
end

-- add time info on the taxi map tooltips
-----------------------------------------------
function InFlight:TaxiNodeOnButtonEnter(button)
-----------------------------------------------
	local index = button:GetID()
	if TaxiNodeGetType(index) == "REACHABLE" then
		local dest = ShortenName(TaxiNodeName(index))
		if vars[source] and vars[source][dest] and vars[source][dest] > 0 then
			gtt:AddLine(format("Time: %s", FormatTime(vars[source][dest])), 1, 1, 1)
		else
			gtt:AddLine("Time: -:--", 0.8, 0.8, 0.8)
		end
		gtt:Show()
	end
end

-- main hook to determine when a flight is taken
------------------------------------
function InFlight:TakeTaxiNode(slot)
------------------------------------
	if TaxiNodeGetType(slot) == "REACHABLE" then
		-- check to see if there's data for the source/destination pair
		destination = ShortenName(TaxiNodeName(slot))
		if vars[source] and vars[source][destination] and vars[source][destination] > 0 then
			endTime = vars[source][destination]
		else
			endTime = nil
		end

		-- confirm or start the flight
		if db.confirmflight then
			self:ConfirmFlight(slot, destination, endTime)
		else
			self:StartTimer(slot)
		end
	end
end

-- setup the timer bar then begin timing
local IsOnFlightYet
----------------------------------
function InFlight:StartTimer(slot)
----------------------------------
	if not f then 
		self:CreateBar() 
	end
	if db.inline then
		locText:SetText(destination)
	else
		locText:SetText(format(locformat, source, destination))
	end
	
	-- start the timers and setup statusbar
	if endTime then
		f:SetMinMaxValues(0, endTime)
		if db.fill == "Deplete" then
			f:SetValue(endTime)
		else
			f:SetValue(0)
		end
		f:SetStatusBarColor(db.barcolor.r, db.barcolor.g, db.barcolor.b, 0.8)
		spark:Show()
		-- values to compute now instead of repeating in onupdate func
		rat = f:GetWidth() / endTime
		timeText2:SetText(" / "..FormatTime(endTime))
	else
		f:SetMinMaxValues(0,1)
		f:SetValue(1)
		f:SetStatusBarColor(db.unknowncolor.r, db.unknowncolor.g, db.unknowncolor.b, 0.8)
		timeText2:SetText(" / ??")
		spark:Hide()
	end
	if db.inline then
		timeText2:SetPoint("LEFT", f, "RIGHT", -6 - timeText2:GetStringWidth(), 0)
	end
	porttaken = nil
	totalTime = 0
	-- getting on taxi is not instant, recheck until detected
	self:ScheduleRepeatingEvent("isactive", IsOnFlightYet, 0.2, self)
	
	if slot then 
		self.hooks.TakeTaxiNode(slot) 
	end
end

-- ask user to confirm flight before taking it
StaticPopupDialogs["INFLIGHT_CONFIRMTAKEOFF"] = {
	text = "",
	button1 = TEXT(OKAY),
	button2 = TEXT(CANCEL),
	OnAccept = function(data) InFlight:StartTimer(data) end,
	timeout = 0,
	exclusive = 1,
	hideOnEscape = 1,
}
-----------------------------------------------------------
function InFlight:ConfirmFlight(slot, location, flighttime)
-----------------------------------------------------------
	local texttime = ""
	if flighttime then
		texttime = format("This flight will take %s.", FormatTime(flighttime))
	end
	StaticPopupDialogs["INFLIGHT_CONFIRMTAKEOFF"].text = TEXT(format("Are you sure you want to fly to %s? %s", location, texttime))
	local dialog = StaticPopup_Show("INFLIGHT_CONFIRMTAKEOFF")
	if dialog then 
		dialog.data = slot 
	end
end

-- record POI location for Cartographer_Notes
------------------------------
function InFlight:NoteIt(misc)
------------------------------
	SetMapToCurrentZone()
	local x,y = GetPlayerMapPosition("player")
	if not Cartographer_Notes:GetNearbyNote(GetRealZoneText(), x, y, 20, "InFlight", true) then
		Cartographer_Notes:SetNote(GetRealZoneText(), x, y, (misc and "Taxi2") or "Taxi", "InFlight",
			"title", "Flight Master", "titleR", 0, "titleB", 0 )
	end
end

-- support for druid-only paths and EPL tower paths, main code in InFlight_Load
---------------------------------------
function InFlight:StartMiscFlight(s, d)
---------------------------------------
	self:TAXIMAP_OPENED(true)
	if vars[s] and vars[s][d] then
		endTime = vars[s][d]
	else
		endTime = nil
	end
	source, destination = s, d
	self:StartTimer()
end


local FlightTimeOnUpdate
do
	-- save data
	local function SaveFlightTime()
		vars[source] = vars[source] or { }
		vars[source][destination] = floor(totalTime + 0.5)
		totalTime = 0
	end
	
	local ontaxi = UnitOnTaxi
	local elapsed
	-- UnitOnTaxi doesn't trigger immediately at flight start, so keep checking
	function IsOnFlightYet() 
		if ontaxi("player") then 
			self:CancelScheduledEvent("isactive")
			elapsed = 0
			f:Show()
		end 
	end
	
	-- in flight onupdate function
	function FlightTimeOnUpdate()
		elapsed = elapsed + arg1
		-- update only 5 times a second
		if elapsed < 0.2 then return end
		
		totalTime = totalTime + elapsed
		elapsed = 0
		 -- flight ended
		if not ontaxi("player") then
			if not porttaken then 
				SaveFlightTime() 
			end
			porttaken = nil
			endTime = nil
			f:Hide()
			return
		end
		-- update statusbar if destination time is known
		if endTime then
			-- in case the flight is longer than expected
			if totalTime - 2 > endTime then 
				f:SetMinMaxValues(0,1)
				f:SetValue(1)
				f:SetStatusBarColor(db.unknowncolor.r, db.unknowncolor.g, db.unknowncolor.b, 0.8)
				timeText2:SetText(" / ??")
				spark:Hide()
				endTime = nil
			else
				local value = (db.fill == "Deplete" and (endTime - totalTime)) or totalTime
				f:SetValue(value)
				spark:SetPoint("CENTER", f, "LEFT", value * rat, 0)
				timeText:SetText(FormatTime(value))
			end
		-- destination time is unknown, so show timing info
		else
			timeText:SetText(FormatTime(totalTime))
		end
	end
end

-- create the statusbar frame for the inflight timer
local font
-----------------------------
function InFlight:CreateBar()
-----------------------------
	f = CreateFrame("StatusBar", "InFlightBar", UIParent)
	f:Hide()
	if not db.x then
		f:SetPoint("CENTER", UIParent, "CENTER", 0, 200)
	else
		f:SetPoint("TOPLEFT", UIParent, "TOPLEFT", db.x, db.y)
	end
	f:SetMovable(true)
	f:EnableMouse(true)
	f:RegisterForDrag("LeftButton")
	f:SetScript("OnDragStart", function() if IsShiftKeyDown() then this:StartMoving() end end)
	f:SetScript("OnDragStop", function() 
		this:StopMovingOrSizing() 
		local a,b,c,d,e = this:GetPoint()
		if a == "TOPLEFT" and c == "TOPLEFT" then
			db.x = floor(d + 0.5)
			db.y = floor(e + 0.5)
		end
	end)
	f:SetScript("OnEnter", function()
		gtt:SetOwner(this, "ANCHOR_RIGHT")
		gtt:SetText("InFlight", 1, 1, 1)
		gtt:AddLine(" <Shift left-click> to move.", 0, 1, 0)
		gtt:AddLine(" <Right-click> for menu.", 0, 1, 0)
		gtt:Show()
	end)
	f:SetScript("OnLeave", function() gtt:Hide() end)
	f:SetScript("OnUpdate", FlightTimeOnUpdate)
	
	f.spark = f:CreateTexture(nil, "OVERLAY")
	f.locText = f:CreateFontString(nil, "OVERLAY")
	f.timeText = f:CreateFontString(nil, "OVERLAY")
	f.timeText2 = f:CreateFontString(nil, "OVERLAY")
	f.bord = CreateFrame("Frame", nil, f)
	spark, locText, timeText, timeText2, bord = f.spark, f.locText, f.timeText, f.timeText2, f.bord
	
	bord:ClearAllPoints()
	bord:SetPoint("TOPLEFT", -5, 5)
	bord:SetPoint("BOTTOMRIGHT", 5, -5)
	bord:SetFrameStrata("LOW")
	
	timeText:SetJustifyH("RIGHT")
	timeText2:SetJustifyH("LEFT")
	locText:SetJustifyV("CENTER")
	timeText:SetPoint("RIGHT", timeText2, "LEFT", 1, 0)
	
	dew:Register( f,
		'children', function() dew:FeedAceOptionsTable(opts) end,
		'cursorX', true, 'cursorY', true )

	font = GameFontHighlightSmall:GetFont()
	
	spark:SetTexture("Interface\\CastingBar\\UI-CastingBar-Spark")
	spark:SetWidth(16)
	spark:SetBlendMode("ADD")
	
	self:UpdateLook()
end

do
	local function UpdateTextLook(t, noclear)
		t:SetFont(font, db.fontsize)
		if not noclear then
			t:ClearAllPoints()
		end
		t:SetShadowColor(0.1, 0.1, 0.1)
		t:SetShadowOffset(1,-1)
		t:SetTextColor(db.fontcolor.r, db.fontcolor.g, db.fontcolor.b)
	end
	local bdrop = {
		edgeSize = 16,
		insets = {},
	}
	local bdi = bdrop.insets
	------------------------------
	function InFlight:UpdateLook()
	------------------------------
		if not f then return end
		
		f:SetWidth(db.width)
		f:SetHeight(db.height)
		
		local w = f:GetWidth()
		local texture = smed:Fetch("statusbar", db.texture)
		
		-- background and border texture settings
		local inset = (db.border == "Textured" and 3) or 5
		bdrop.bgFile = texture
		bdrop.edgeFile = (db.border=="Default" and "Interface\\Tooltips\\UI-Tooltip-Border") or ""
		bdi.left, bdi.right, bdi.top, bdi.bottom = inset, inset, inset, inset
		bord:SetBackdrop(bdrop)
		bord:SetBackdropColor(db.backcolor.r, db.backcolor.g, db.backcolor.b, 0.6)
		bord:SetBackdropBorderColor(db.bordercolor.r, db.bordercolor.g, db.bordercolor.b, 0.8)
		
		-- statusbar texture
		f:SetStatusBarTexture(texture)
		
		-- in case we're in flight
		if endTime then
			rat = w / endTime
		end
		
		-- spark
		spark:SetHeight(db.height*2.44)
		
		-- update text
		UpdateTextLook(locText)
		UpdateTextLook(timeText, true)
		UpdateTextLook(timeText2)
		timeText:SetWidth(w*0.4)
		timeText2:SetWidth(w*0.4)
		if db.inline then
			timeText2:SetPoint("LEFT", f, "RIGHT", -6 - timeText2:GetStringWidth(), 0)

			locText:SetJustifyH("LEFT")
			locText:SetWidth(w*0.7)
			locText:SetPoint("LEFT", f, "LEFT", 6, 0)	
		else
			timeText2:SetPoint("LEFT", f, "CENTER", -3, 0)

			locText:SetJustifyH("CENTER")
			locText:SetWidth(max(200, w*1.2))
			locText:SetPoint("BOTTOM", f, "TOP", 0, (db.border~="Default" and 1) or 3)
		end	
	end
end

-------------------------------
function InFlight:ShowOptions()
-------------------------------
	dew:Open( UIParent, 
	          'children', function() dew:FeedAceOptionsTable(opts) end, 
	          'cursorX', true, 'cursorY', true ) 
end

-------------------------------
function InFlight:LoadOptions()
-------------------------------
	opts = {
		type = "group",
		args = {
			header = {
				type = "header", name = "InFlight",
				order = 1,
			},
			frame = {
				type = "group", name = "Appearance", desc = "Change the way the statusbar looks and behaves.",
				order = 2,
				args = {
					inline = {
						type = "toggle", name = "Inline Mode", desc = "Compact text inside the statusbar",
						set = function(v) 
							db.inline = v 
							self:UpdateLook() 
							if destination then
								locText:SetText((db.inline and destination) or format(locformat, source, destination))
							end
						end,
						get = function() return db.inline end,
						order = 1,
					},
					barfill = {
						type = "text", name = "Bar Fill", desc = "Set the fill type of the statusbar",
						set = function(v) db.fill = v self:UpdateLook() end,
						get = function() return db.fill end,
						validate = { "Fill up", "Deplete", },
						order = 2,
					},
					height = {
						type = "range", name = "Height", desc = "Height",
						set = function(v) db.height = v self:UpdateLook() end,
						get = function() return db.height end,
						min = 4, max = 100, step = 2,
						order = 3,
					},
					width = {
						type = "range", name = "Width", desc = "Width",
						set = function(v) db.width = v self:UpdateLook() end,
						get = function() return db.width end,
						min = 100, max = 600, step = 5,
						order = 4,
					},
					fontsize = {
						type = "range", name = "Font Size", desc = "Font Size",
						set = function(v) db.fontsize = v self:UpdateLook() end,
						get = function() return db.fontsize end,
						min = 4, max = 40, step = 1,
						order = 5,
					},
					border = {
						type = "text", name = "Border", desc = "Border Options",
						set = function(v) db.border = v self:UpdateLook() end,
						get = function() return db.border end,
						validate = { "None", "Default", "Textured", },
						order = 6,
					},
					bartexture = {
						type = "text", name = "Bar Texture", desc = "Set the texture of the statusbar",
						get = function() return db.texture end,
						set = function(v) 
							db.texture = v 
							f:SetStatusBarTexture(smed:Fetch("statusbar", db.texture)) 
							self:UpdateLook()
						end,
						validate = smed:List("statusbar"),
						order = 7,
					},
					barcolor = {
						type = "color", name = "Bar Color", desc = "Set the color of the fill bar.",
						get = function() return db.barcolor.r, db.barcolor.g, db.barcolor.b end,
						set = function(r,g,b) 
							db.barcolor = {r=r, g=g, b=b} 
							if endTime then 
								f:SetStatusBarColor(r, g, b) 
							end
						end,
						order = 8,
					},
					unknowncolor = {
						type = "color", name = "Unknown Color", desc = "Set the color of the fill bar for unknown flights.",
						get = function() return db.unknowncolor.r, db.unknowncolor.g, db.unknowncolor.b end,
						set = function(r,g,b) 
							db.unknowncolor = {r=r, g=g, b=b} 
							if not endTime then
								f:SetStatusBarColor(r, g, b) 
							end
						end,
						order = 9,
					},
					bordercolor = {
						type = "color", name = "Border Color", desc = "Sets the color of the statusbar's border.",
						get = function() return db.bordercolor.r, db.bordercolor.g, db.bordercolor.b end,
						set = function(r,g,b) db.bordercolor = {r=r,g=g,b=b} self:UpdateLook() end,
						order = 10,
					},
					backcolor = {
						type = "color", name = "Background Color", desc = "Sets the color of the statusbar's background.",
						get = function() return db.backcolor.r, db.backcolor.g, db.backcolor.b end,
						set = function(r,g,b) db.backcolor = {r=r,g=g,b=b} self:UpdateLook() end,
						order = 11,
					},
					fontcolor = {
						type = "color", name = "Font Color", desc = "Sets the color of the text.",
						get = function() return db.fontcolor.r, db.fontcolor.g, db.fontcolor.b end,
						set = function(r,g,b) db.fontcolor = {r=r,g=g,b=b} self:UpdateLook() end,
						order = 12,
					},
				},
			},
			dismount = {
				type = "toggle", name = "Auto-dismount", desc = "Auto-dismount when talking to a flightmaster.",
				set = function(v) db.dismount = v end,
				get = function() return db.dismount end,
				order = 4,
			},
			confirm = {
				type = "toggle", name = "Confirm flight", desc = "Confirm the flight before taking it.",
				set = function(v) db.confirmflight = v end,
				get = function() return db.confirmflight end,
				order = 5,
			},
		},
	}
end